#ifndef _KGR_PATCH_BSC1021417_LOCAL_H
#define _KGR_PATCH_BSC1021417_LOCAL_H

#include <linux/skbuff.h>
#include <linux/interrupt.h>
#include <linux/if_vlan.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include <xen/interface/io/netif.h>

#define MAX_PENDING_REQS (1U << CONFIG_XEN_NETDEV_TX_SHIFT)
#define MAX_MFN_ALLOC 64
#define NET_TX_RING_SIZE __CONST_RING_SIZE(netif_tx, PAGE_SIZE)
#define NET_RX_RING_SIZE __CONST_RING_SIZE(netif_rx, PAGE_SIZE)
#define MASK_PEND_IDX(_i) ((_i)&(MAX_PENDING_REQS-1))
#define netif_get(_b) (atomic_inc(&(_b)->refcnt))
#define netif_put(_b)						\
	do {							\
		if ( atomic_dec_and_test(&(_b)->refcnt) )	\
			wake_up(&(_b)->waiting_to_free);	\
	} while (0)
#define PKT_PROT_LEN    (ETH_HLEN + VLAN_HLEN + \
			 sizeof(struct iphdr) + MAX_IPOPTLEN + \
			 sizeof(struct tcphdr) + MAX_TCP_OPTION_SPACE)
#define INVALID_PENDING_IDX 0xffff

typedef unsigned int RING_IDX;
typedef unsigned int pending_ring_idx_t;

enum {
	NETBK_DONT_COPY_SKB,
	NETBK_DELAYED_COPY_SKB,
	NETBK_ALWAYS_COPY_SKB,
};

typedef struct netif_st {
	/* Unique identifier for this interface. */
	domid_t          domid;
	unsigned int     group;
	unsigned int     handle;

	u8               fe_dev_addr[6];

	unsigned int     irq;

	struct netif_st *notify_link[2];
#define RX_IDX 0
#define TX_IDX 1
#define rx_notify_link notify_link[RX_IDX]
#define tx_notify_link notify_link[TX_IDX]

	/* The shared rings and indexes. */
	netif_tx_back_ring_t tx;
	netif_rx_back_ring_t rx;
	struct vm_struct *tx_comms_area;
	struct vm_struct *rx_comms_area;

	/* Flags that must not be set in dev->features */
	int features_disabled;

	/* Frontend feature information. */
	u8 can_sg:1;
	u8 gso:1;
	u8 csum:1;

	/* Internal feature information. */
	u8 can_queue:1;	/* can queue packets for receiver? */
	u8 copying_receiver:1;	/* copy packets to receiver?       */

	u8 busted:1;

	/* Allow netif_be_start_xmit() to peek ahead in the rx request ring. */
	RING_IDX rx_req_cons_peek;

	/* Transmit shaping: allow 'credit_bytes' every 'credit_usec'. */
	unsigned long   credit_bytes;
	unsigned long   credit_usec;
	unsigned long   remaining_credit;
	struct timer_list credit_timeout;
	u64 credit_window_start;

	/* Enforce draining of the transmit queue. */
	struct timer_list tx_queue_timeout;

	/* Statistics */
	unsigned long nr_copied_rx_skbs;
	unsigned long nr_copied_tx_skbs;
	unsigned long nr_coalesced_skbs;
	unsigned long rx_gso_csum_fixups;

	/* Miscellaneous private stuff. */
	struct list_head list;  /* scheduling list */
	atomic_t         refcnt;
	struct net_device *dev;

	unsigned int carrier;

	wait_queue_head_t waiting_to_free;
} netif_t;

struct netbk_tx_gop {
	gnttab_map_grant_ref_t *map;
	gnttab_copy_t *copy;
	union {
		void *ptr;
		struct {
			netif_t *head, *tail;
		} notify;
	};
};

struct pending_tx_info {
	netif_tx_request_t req;
	grant_handle_t grant_handle;
	netif_t *netif;
};

struct netbk_tx_pending_inuse {
	struct list_head list;
	unsigned long alloc_time;
};

struct netbk_rx_meta {
	skb_frag_t frag;
	u16 id;
	u8 copy;
	u8 tail;
};

struct xen_netbk {
	atomic_t nr_groups;

	struct {
		pending_ring_idx_t pending_prod, pending_cons;
		pending_ring_idx_t dealloc_prod, dealloc_cons;
		struct sk_buff_head queue;
		struct tasklet_struct tasklet;
		struct list_head schedule_list;
		spinlock_t schedule_list_lock;
		spinlock_t release_lock;
		struct page **mmap_pages;
		struct timer_list pending_timer;
		struct list_head pending_inuse_head;
		struct pending_tx_info pending_info[MAX_PENDING_REQS];
		struct netbk_tx_pending_inuse pending_inuse[MAX_PENDING_REQS];
		u16 pending_ring[MAX_PENDING_REQS];
		u16 dealloc_ring[MAX_PENDING_REQS];
		union {
			gnttab_map_grant_ref_t map_ops[MAX_PENDING_REQS];
			gnttab_unmap_grant_ref_t unmap_ops[MAX_PENDING_REQS];
			gnttab_copy_t copy_ops[2 * MAX_PENDING_REQS - 1];
			multicall_entry_t mcl[0];
		};
		gnttab_copy_t copy_op;
		netif_tx_request_t slots[XEN_NETIF_NR_SLOTS_MIN];
	} tx;

	wait_queue_head_t action_wq;
	struct task_struct *task;

	struct xen_netbk_rx {
		struct sk_buff_head queue;
		struct tasklet_struct tasklet;
		struct timer_list timer;
		unsigned int alloc_index;
		struct multicall_entry mcl[NET_RX_RING_SIZE+3];
		struct mmu_update mmu[NET_RX_RING_SIZE];
		/*
		 * An skb may have a maximal number of frags but still be
		 * small in overall size. Thus the worst case number of copy
		 * operations is one more (for the head) than MAX_SKB_FRAGS
		 * per ring slot.
		 */
		struct gnttab_copy grant_copy_op[(MAX_SKB_FRAGS + 1) *
						 NET_RX_RING_SIZE];
		struct gnttab_transfer grant_trans_op;
		struct netbk_rx_meta meta[NET_RX_RING_SIZE];
		unsigned long mfn_list[MAX_MFN_ALLOC];
	} rx;
};

struct netbk_tx_cb {
	u16 copy_slots;
	u16 pending_idx[1 + XEN_NETIF_NR_SLOTS_MIN];
};
#define netbk_tx_cb(skb) ((struct netbk_tx_cb *)skb->cb)

#endif
